"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const identifierRegex = /^[a-zA-Z_$][a-zA-Z_$0-9]*$/;
/*
  If the part is a function, we need to get the actual value of the parameters.
  A parameter x or y in func(x, y) may be
    a) Assigned a value in one of the dicts
    b) OR should be treated as a string literal "x" or "y"
*/
function getArgumentValue(dicts) {
    return (a) => a === "true" || a === "false"
        ? { value: a === "true" }
        : isNaN(Number(a))
            ? identifierRegex.test(a)
                ? (() => {
                    for (const dict of dicts) {
                        if (typeof dict === "function") {
                            const val = dict(a);
                            if (val) {
                                return val;
                            }
                        }
                        else {
                            if (hasOwnProperty(dict, a)) {
                                return { value: dict[a] };
                            }
                        }
                    }
                    return { value: a };
                })()
                : { error: `${a} is not a valid identifier.` }
            : { value: +a };
}
/*
    Split path expression into
      a) objects
      b) functions and their parameters
*/
function analyzePath(encodedPath, dicts) {
    //The path would be url encoded
    const path = decodeURI(encodedPath);
    const parts = path.indexOf(".") ? path.split(".") : [path];
    return parts.map(part => {
        const openingBracket = part.indexOf("(");
        const isFunction = openingBracket > -1;
        return isFunction
            ? (() => {
                const closingBracket = part.indexOf(")");
                const identifier = part.substring(0, openingBracket);
                const argsString = part.substring(openingBracket + 1, closingBracket);
                const args = argsString
                    .split(",")
                    .map(a => a.trim())
                    .map(getArgumentValue(dicts));
                const result = { type: "function", identifier, args };
                return result;
            })()
            : { type: "object", identifier: part };
    });
}
class RoutingError {
    constructor(message, type) {
        this.message = message;
        this.type = type;
    }
}
exports.RoutingError = RoutingError;
function isIterable(gen) {
    return ((gen[Symbol.iterator] && typeof gen[Symbol.iterator] === "function") ||
        (Symbol.asyncIterator &&
            gen[Symbol.asyncIterator] &&
            typeof gen[Symbol.asyncIterator] === "function"));
}
function __isIterable(gen) {
    return gen.next && typeof gen.next === "function";
}
function hasOwnProperty(obj, prop) {
    return (typeof obj === "object" && Object.prototype.hasOwnProperty.call(obj, prop));
}
function route(_app, _expression, dicts = [], _options = {}) {
    return __awaiter(this, void 0, void 0, function* () {
        const options = Object.assign({}, _options, { index: _options.index || "index" });
        function iterateToEnd(resultOrGenerator, current) {
            return __awaiter(this, void 0, void 0, function* () {
                if (__isIterable(resultOrGenerator)) {
                    const gen = resultOrGenerator;
                    while (true) {
                        const nextVal = yield gen.next();
                        if (options.onNextValue && !nextVal.done) {
                            options.onNextValue(yield nextVal.value, current);
                        }
                        if (nextVal.done) {
                            return yield nextVal.value;
                        }
                    }
                }
                else {
                    return resultOrGenerator;
                }
            });
        }
        const app = typeof _app === "function" ? _app() : _app;
        const expression = options.useSlash
            ? _expression.replace(/\//g, ".")
            : _expression;
        const additionalArgs = options.args || [];
        const parts = expression ? analyzePath(expression, dicts) : [];
        let obj, error, current = app;
        for (const part of parts) {
            obj = obj ? `${obj}.${part.identifier}` : `${part.identifier}`;
            if (typeof current !== "undefined") {
                if (options.modifyHandler) {
                    current = options.modifyHandler(part.identifier, current);
                }
                if (hasOwnProperty(current, part.identifier)) {
                    if (part.type === "function") {
                        const fn = current[part.identifier];
                        if (typeof fn === "function") {
                            const resultOrGenerator = yield fn.apply(current, options.prependArgs
                                ? additionalArgs.concat(part.args.map(a => a.value))
                                : part.args.map(a => a.value).concat(additionalArgs));
                            current = yield iterateToEnd(resultOrGenerator, current);
                        }
                        else if (typeof fn === "undefined") {
                            error = new RoutingError("The requested path was not found.", "NOT_FOUND");
                            break;
                        }
                        else {
                            const ref = current[part.identifier];
                            const resultOrGenerator = yield (typeof ref === "function"
                                ? ref.apply(current, additionalArgs)
                                : ref);
                            current = yield iterateToEnd(resultOrGenerator, current);
                        }
                    }
                    else {
                        const ref = current[part.identifier];
                        const resultOrGenerator = yield (typeof ref === "function"
                            ? ref.apply(current, additionalArgs)
                            : ref);
                        current = yield iterateToEnd(resultOrGenerator, current);
                    }
                }
                else {
                    error = new RoutingError("The requested path was not found.", "NOT_FOUND");
                    break;
                }
            }
            else {
                break;
            }
        }
        const finalResult = error
            ? error
            : hasOwnProperty(current, options.index)
                ? yield (() => __awaiter(this, void 0, void 0, function* () {
                    if (options.modifyHandler) {
                        current = options.modifyHandler(options.index, current);
                    }
                    const resultOrGenerator = typeof current[options.index] === "function"
                        ? current[options.index].apply(current, additionalArgs)
                        : current[options.index];
                    return yield iterateToEnd(resultOrGenerator, current);
                }))()
                : current;
        return yield finalResult;
    });
}
exports.default = route;
//# sourceMappingURL=index.js.map